#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Sheets API hook."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from google.api_core.client_options import ClientOptions
from googleapiclient.discovery import build

from airflow.exceptions import AirflowException
from airflow.providers.google.common.hooks.base_google import GoogleBaseHook


class GSheetsHook(GoogleBaseHook):
    """
    Interact with Google Sheets via Google Cloud connection.

    Reading and writing cells in Google Sheet: https://developers.google.com/sheets/api/guides/values

    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param api_version: API Version
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account.
    :param api_endpoint: Optional. Custom API endpoint, i.e: regional or private endpoint.
        This can be used to target private VPC or restricted access endpoints.
    """

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        api_version: str = "v4",
        impersonation_chain: str | Sequence[str] | None = None,
        api_endpoint: str | None = None,
    ) -> None:
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            impersonation_chain=impersonation_chain,
        )
        self.gcp_conn_id = gcp_conn_id
        self.api_version = api_version
        self.api_endpoint = api_endpoint
        self._conn = None

    def get_conn(self) -> Any:
        """
        Retrieve connection to Google Sheets.

        :return: Google Sheets services object.
        """
        if not self._conn:
            http_authorized = self._authorize()
            client_options = None
            if self.api_endpoint:
                client_options = ClientOptions(api_endpoint=self.api_endpoint)
            self._conn = build(
                "sheets",
                self.api_version,
                http=http_authorized,
                cache_discovery=False,
                client_options=client_options,
            )

        return self._conn

    def get_values(
        self,
        spreadsheet_id: str,
        range_: str,
        major_dimension: str = "DIMENSION_UNSPECIFIED",
        value_render_option: str = "FORMATTED_VALUE",
        date_time_render_option: str = "SERIAL_NUMBER",
    ) -> list:
        """
        Get values from Google Sheet from a single range.

        https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/get

        :param spreadsheet_id: The Google Sheet ID to interact with
        :param range_: The A1 notation of the values to retrieve.
        :param major_dimension: Indicates which dimension an operation should apply to.
            DIMENSION_UNSPECIFIED, ROWS, or COLUMNS
        :param value_render_option: Determines how values should be rendered in the output.
            FORMATTED_VALUE, UNFORMATTED_VALUE, or FORMULA
        :param date_time_render_option: Determines how dates should be rendered in the output.
            SERIAL_NUMBER or FORMATTED_STRING
        :return: An array of sheet values from the specified sheet.
        """
        service = self.get_conn()

        response = (
            service.spreadsheets()
            .values()
            .get(
                spreadsheetId=spreadsheet_id,
                range=range_,
                majorDimension=major_dimension,
                valueRenderOption=value_render_option,
                dateTimeRenderOption=date_time_render_option,
            )
            .execute(num_retries=self.num_retries)
        )

        return response.get("values", [])

    def batch_get_values(
        self,
        spreadsheet_id: str,
        ranges: list,
        major_dimension: str = "DIMENSION_UNSPECIFIED",
        value_render_option: str = "FORMATTED_VALUE",
        date_time_render_option: str = "SERIAL_NUMBER",
    ) -> dict:
        """
        Get values from Google Sheet from a list of ranges.

        https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/batchGet

        :param spreadsheet_id: The Google Sheet ID to interact with
        :param ranges: The A1 notation of the values to retrieve.
        :param major_dimension: Indicates which dimension an operation should apply to.
            DIMENSION_UNSPECIFIED, ROWS, or COLUMNS
        :param value_render_option: Determines how values should be rendered in the output.
            FORMATTED_VALUE, UNFORMATTED_VALUE, or FORMULA
        :param date_time_render_option: Determines how dates should be rendered in the output.
            SERIAL_NUMBER or FORMATTED_STRING
        :return: Google Sheets API response.
        """
        service = self.get_conn()

        response = (
            service.spreadsheets()
            .values()
            .batchGet(
                spreadsheetId=spreadsheet_id,
                ranges=ranges,
                majorDimension=major_dimension,
                valueRenderOption=value_render_option,
                dateTimeRenderOption=date_time_render_option,
            )
            .execute(num_retries=self.num_retries)
        )

        return response

    def update_values(
        self,
        spreadsheet_id: str,
        range_: str,
        values: list,
        major_dimension: str = "ROWS",
        value_input_option: str = "RAW",
        include_values_in_response: bool = False,
        value_render_option: str = "FORMATTED_VALUE",
        date_time_render_option: str = "SERIAL_NUMBER",
    ) -> dict:
        """
        Update values from Google Sheet from a single range.

        https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/update

        :param spreadsheet_id: The Google Sheet ID to interact with.
        :param range_: The A1 notation of the values to retrieve.
        :param values: Data within a range of the spreadsheet.
        :param major_dimension: Indicates which dimension an operation should apply to.
            DIMENSION_UNSPECIFIED, ROWS, or COLUMNS
        :param value_input_option: Determines how input data should be interpreted.
            RAW or USER_ENTERED
        :param include_values_in_response: Determines if the update response should
            include the values of the cells that were updated.
        :param value_render_option: Determines how values should be rendered in the output.
            FORMATTED_VALUE, UNFORMATTED_VALUE, or FORMULA
        :param date_time_render_option: Determines how dates should be rendered in the output.
            SERIAL_NUMBER or FORMATTED_STRING
        :return: Google Sheets API response.
        """
        service = self.get_conn()
        body = {"range": range_, "majorDimension": major_dimension, "values": values}

        response = (
            service.spreadsheets()
            .values()
            .update(
                spreadsheetId=spreadsheet_id,
                range=range_,
                valueInputOption=value_input_option,
                includeValuesInResponse=include_values_in_response,
                responseValueRenderOption=value_render_option,
                responseDateTimeRenderOption=date_time_render_option,
                body=body,
            )
            .execute(num_retries=self.num_retries)
        )

        return response

    def batch_update_values(
        self,
        spreadsheet_id: str,
        ranges: list,
        values: list,
        major_dimension: str = "ROWS",
        value_input_option: str = "RAW",
        include_values_in_response: bool = False,
        value_render_option: str = "FORMATTED_VALUE",
        date_time_render_option: str = "SERIAL_NUMBER",
    ) -> dict:
        """
        Update values from Google Sheet for multiple ranges.

        https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/batchUpdate

        :param spreadsheet_id: The Google Sheet ID to interact with
        :param ranges: The A1 notation of the values to retrieve.
        :param values: Data within a range of the spreadsheet.
        :param major_dimension: Indicates which dimension an operation should apply to.
            DIMENSION_UNSPECIFIED, ROWS, or COLUMNS
        :param value_input_option: Determines how input data should be interpreted.
            RAW or USER_ENTERED
        :param include_values_in_response: Determines if the update response should
            include the values of the cells that were updated.
        :param value_render_option: Determines how values should be rendered in the output.
            FORMATTED_VALUE, UNFORMATTED_VALUE, or FORMULA
        :param date_time_render_option: Determines how dates should be rendered in the output.
            SERIAL_NUMBER or FORMATTED_STRING
        :return: Google Sheets API response.
        """
        if len(ranges) != len(values):
            raise AirflowException(
                f"'Ranges' and 'Lists' must be of equal length. "
                f"'Ranges' is of length: {len(ranges)} and 'Values' is of length: {len(values)}."
            )
        service = self.get_conn()
        data = []
        for idx, range_ in enumerate(ranges):
            value_range = {"range": range_, "majorDimension": major_dimension, "values": values[idx]}
            data.append(value_range)
        body = {
            "valueInputOption": value_input_option,
            "data": data,
            "includeValuesInResponse": include_values_in_response,
            "responseValueRenderOption": value_render_option,
            "responseDateTimeRenderOption": date_time_render_option,
        }

        response = (
            service.spreadsheets()
            .values()
            .batchUpdate(spreadsheetId=spreadsheet_id, body=body)
            .execute(num_retries=self.num_retries)
        )

        return response

    def append_values(
        self,
        spreadsheet_id: str,
        range_: str,
        values: list,
        major_dimension: str = "ROWS",
        value_input_option: str = "RAW",
        insert_data_option: str = "OVERWRITE",
        include_values_in_response: bool = False,
        value_render_option: str = "FORMATTED_VALUE",
        date_time_render_option: str = "SERIAL_NUMBER",
    ) -> dict:
        """
        Append values from Google Sheet from a single range.

        https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/append

        :param spreadsheet_id: The Google Sheet ID to interact with
        :param range_: The A1 notation of the values to retrieve.
        :param values: Data within a range of the spreadsheet.
        :param major_dimension: Indicates which dimension an operation should apply to.
            DIMENSION_UNSPECIFIED, ROWS, or COLUMNS
        :param value_input_option: Determines how input data should be interpreted.
            RAW or USER_ENTERED
        :param insert_data_option: Determines how existing data is changed when new data is input.
            OVERWRITE or INSERT_ROWS
        :param include_values_in_response: Determines if the update response should
            include the values of the cells that were updated.
        :param value_render_option: Determines how values should be rendered in the output.
            FORMATTED_VALUE, UNFORMATTED_VALUE, or FORMULA
        :param date_time_render_option: Determines how dates should be rendered in the output.
            SERIAL_NUMBER or FORMATTED_STRING
        :return: Google Sheets API response.
        """
        service = self.get_conn()
        body = {"range": range_, "majorDimension": major_dimension, "values": values}

        response = (
            service.spreadsheets()
            .values()
            .append(
                spreadsheetId=spreadsheet_id,
                range=range_,
                valueInputOption=value_input_option,
                insertDataOption=insert_data_option,
                includeValuesInResponse=include_values_in_response,
                responseValueRenderOption=value_render_option,
                responseDateTimeRenderOption=date_time_render_option,
                body=body,
            )
            .execute(num_retries=self.num_retries)
        )

        return response

    def clear(self, spreadsheet_id: str, range_: str) -> dict:
        """
        Clear values from Google Sheet from a single range.

        https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/clear

        :param spreadsheet_id: The Google Sheet ID to interact with
        :param range_: The A1 notation of the values to retrieve.
        :return: Google Sheets API response.
        """
        service = self.get_conn()

        response = (
            service.spreadsheets()
            .values()
            .clear(spreadsheetId=spreadsheet_id, range=range_)
            .execute(num_retries=self.num_retries)
        )

        return response

    def batch_clear(self, spreadsheet_id: str, ranges: list) -> dict:
        """
        Clear values from Google Sheet from a list of ranges.

        https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/batchClear

        :param spreadsheet_id: The Google Sheet ID to interact with
        :param ranges: The A1 notation of the values to retrieve.
        :return: Google Sheets API response.
        """
        service = self.get_conn()
        body = {"ranges": ranges}

        response = (
            service.spreadsheets()
            .values()
            .batchClear(spreadsheetId=spreadsheet_id, body=body)
            .execute(num_retries=self.num_retries)
        )

        return response

    def get_spreadsheet(self, spreadsheet_id: str):
        """
        Retrieve spreadsheet matching the given id.

        :param spreadsheet_id: The spreadsheet id.
        :return: An spreadsheet that matches the sheet filter.
        """
        response = (
            self.get_conn()
            .spreadsheets()
            .get(spreadsheetId=spreadsheet_id)
            .execute(num_retries=self.num_retries)
        )
        return response

    def get_sheet_titles(self, spreadsheet_id: str, sheet_filter: list[str] | None = None):
        """
        Retrieve the sheet titles from a spreadsheet matching the given id and sheet filter.

        :param spreadsheet_id: The spreadsheet id.
        :param sheet_filter: List of sheet title to retrieve from sheet.
        :return: An list of sheet titles from the specified sheet that match
            the sheet filter.
        """
        response = self.get_spreadsheet(spreadsheet_id=spreadsheet_id)

        if sheet_filter:
            titles = [
                sh["properties"]["title"]
                for sh in response["sheets"]
                if sh["properties"]["title"] in sheet_filter
            ]
        else:
            titles = [sh["properties"]["title"] for sh in response["sheets"]]
        return titles

    def create_spreadsheet(self, spreadsheet: dict[str, Any]) -> dict[str, Any]:
        """
        Create a spreadsheet, returning the newly created spreadsheet.

        :param spreadsheet: an instance of Spreadsheet
            https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets#Spreadsheet
        :return: An spreadsheet object.
        """
        self.log.info("Creating spreadsheet: %s", spreadsheet["properties"]["title"])

        response = (
            self.get_conn().spreadsheets().create(body=spreadsheet).execute(num_retries=self.num_retries)
        )
        self.log.info("Spreadsheet: %s created", spreadsheet["properties"]["title"])
        return response



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from datetime import datetime
from typing import cast

from google.protobuf.field_mask_pb2 import FieldMask

from airflow.models.dag import DAG
from airflow.models.xcom_arg import XComArg
from airflow.providers.google.cloud.operators.workflows import (
    WorkflowsCancelExecutionOperator,
    WorkflowsCreateExecutionOperator,
    WorkflowsCreateWorkflowOperator,
    WorkflowsDeleteWorkflowOperator,
    WorkflowsGetExecutionOperator,
    WorkflowsGetWorkflowOperator,
    WorkflowsListExecutionsOperator,
    WorkflowsListWorkflowsOperator,
    WorkflowsUpdateWorkflowOperator,
)
from airflow.providers.google.cloud.sensors.workflows import WorkflowExecutionSensor

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")

DAG_ID = "cloud_workflows"

LOCATION = "us-central1"
WORKFLOW_ID = f"workflow-{DAG_ID}-{ENV_ID}".replace("_", "-")

# [START how_to_define_workflow]
WORKFLOW_CONTENT = """
- getLanguage:
    assign:
        - inputLanguage: "English"
- readWikipedia:
    call: http.get
    args:
        url: https://www.wikipedia.org/
        query:
            action: opensearch
            search: ${inputLanguage}
    result: wikiResult
- returnResult:
    return: ${wikiResult}
"""

WORKFLOW = {
    "description": "Test workflow",
    "labels": {"airflow-version": "dev"},
    "source_contents": WORKFLOW_CONTENT,
}
# [END how_to_define_workflow]

EXECUTION = {"argument": ""}

SLEEP_WORKFLOW_ID = f"sleep-workflow-{DAG_ID}-{ENV_ID}".replace("_", "-")
SLEEP_WORKFLOW_CONTENT = """
- someSleep:
    call: sys.sleep
    args:
        seconds: 200
"""

SLEEP_WORKFLOW = {
    "description": "Test workflow",
    "labels": {"airflow-version": "dev"},
    "source_contents": SLEEP_WORKFLOW_CONTENT,
}


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "workflows"],
) as dag:
    # [START how_to_create_workflow]
    create_workflow = WorkflowsCreateWorkflowOperator(
        task_id="create_workflow",
        location=LOCATION,
        project_id=PROJECT_ID,
        workflow=WORKFLOW,
        workflow_id=WORKFLOW_ID,
    )
    # [END how_to_create_workflow]

    # [START how_to_update_workflow]
    update_workflow = WorkflowsUpdateWorkflowOperator(
        task_id="update_workflow",
        location=LOCATION,
        project_id=PROJECT_ID,
        workflow_id=WORKFLOW_ID,
        update_mask=FieldMask(paths=["name", "description"]),
    )
    # [END how_to_update_workflow]

    # [START how_to_get_workflow]
    get_workflow = WorkflowsGetWorkflowOperator(
        task_id="get_workflow", location=LOCATION, project_id=PROJECT_ID, workflow_id=WORKFLOW_ID
    )
    # [END how_to_get_workflow]

    # [START how_to_list_workflows]
    list_workflows = WorkflowsListWorkflowsOperator(
        task_id="list_workflows",
        location=LOCATION,
        project_id=PROJECT_ID,
    )
    # [END how_to_list_workflows]

    # [START how_to_delete_workflow]
    delete_workflow = WorkflowsDeleteWorkflowOperator(
        task_id="delete_workflow",
        location=LOCATION,
        project_id=PROJECT_ID,
        workflow_id=WORKFLOW_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END how_to_delete_workflow]

    # [START how_to_create_execution]
    create_execution = WorkflowsCreateExecutionOperator(
        task_id="create_execution",
        location=LOCATION,
        project_id=PROJECT_ID,
        execution=EXECUTION,
        workflow_id=WORKFLOW_ID,
    )
    # [END how_to_create_execution]

    create_execution_id = cast("str", XComArg(create_execution, key="execution_id"))

    # [START how_to_wait_for_execution]
    wait_for_execution = WorkflowExecutionSensor(
        task_id="wait_for_execution",
        location=LOCATION,
        project_id=PROJECT_ID,
        workflow_id=WORKFLOW_ID,
        execution_id=create_execution_id,
    )
    # [END how_to_wait_for_execution]

    # [START how_to_get_execution]
    get_execution = WorkflowsGetExecutionOperator(
        task_id="get_execution",
        location=LOCATION,
        project_id=PROJECT_ID,
        workflow_id=WORKFLOW_ID,
        execution_id=create_execution_id,
    )
    # [END how_to_get_execution]

    # [START how_to_list_executions]
    list_executions = WorkflowsListExecutionsOperator(
        task_id="list_executions", location=LOCATION, project_id=PROJECT_ID, workflow_id=WORKFLOW_ID
    )
    # [END how_to_list_executions]

    create_workflow_for_cancel = WorkflowsCreateWorkflowOperator(
        task_id="create_workflow_for_cancel",
        location=LOCATION,
        project_id=PROJECT_ID,
        workflow=SLEEP_WORKFLOW,
        workflow_id=SLEEP_WORKFLOW_ID,
    )

    create_execution_for_cancel = WorkflowsCreateExecutionOperator(
        task_id="create_execution_for_cancel",
        location=LOCATION,
        project_id=PROJECT_ID,
        execution=EXECUTION,
        workflow_id=SLEEP_WORKFLOW_ID,
    )

    cancel_execution_id = cast("str", XComArg(create_execution_for_cancel, key="execution_id"))

    # [START how_to_cancel_execution]
    cancel_execution = WorkflowsCancelExecutionOperator(
        task_id="cancel_execution",
        location=LOCATION,
        project_id=PROJECT_ID,
        workflow_id=SLEEP_WORKFLOW_ID,
        execution_id=cancel_execution_id,
    )
    # [END how_to_cancel_execution]

    delete_workflow_for_cancel = WorkflowsDeleteWorkflowOperator(
        task_id="delete_workflow_for_cancel",
        location=LOCATION,
        project_id=PROJECT_ID,
        workflow_id=SLEEP_WORKFLOW_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    create_workflow >> update_workflow >> [get_workflow, list_workflows]
    update_workflow >> [create_execution, create_execution_for_cancel]

    wait_for_execution >> [get_execution, list_executions]
    (
        create_workflow_for_cancel
        >> create_execution_for_cancel
        >> cancel_execution
        >> delete_workflow_for_cancel
    )

    [cancel_execution, list_executions] >> delete_workflow

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG for Google Cloud Managed Service for Apache Kafka testing Topic operations.
"""

from __future__ import annotations

import json
import logging
import os
import random
from datetime import datetime
from typing import Any

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.dag import DAG
from airflow.providers.apache.kafka.operators.consume import ConsumeFromTopicOperator
from airflow.providers.apache.kafka.operators.produce import ProduceToTopicOperator
from airflow.providers.google.cloud.operators.managed_kafka import (
    ManagedKafkaCreateClusterOperator,
    ManagedKafkaCreateTopicOperator,
    ManagedKafkaDeleteClusterOperator,
    ManagedKafkaDeleteConsumerGroupOperator,
    ManagedKafkaGetConsumerGroupOperator,
    ManagedKafkaListConsumerGroupsOperator,
    ManagedKafkaUpdateConsumerGroupOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from tests_common.test_utils.api_client_helpers import create_airflow_connection

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
DAG_ID = "managed_kafka_consumer_group_operations"
LOCATION = "us-central1"

CLUSTER_ID = f"cluster_{DAG_ID}_{ENV_ID}".replace("_", "-")
CLUSTER_CONF = {
    "gcp_config": {
        "access_config": {
            "network_configs": [
                {"subnet": f"projects/{PROJECT_ID}/regions/{LOCATION}/subnetworks/default"},
            ],
        },
    },
    "capacity_config": {
        "vcpu_count": 3,
        "memory_bytes": 3221225472,
    },
}
TOPIC_ID = f"topic_{DAG_ID}_{ENV_ID}".replace("_", "-")
TOPIC_CONF = {
    "partition_count": 3,
    "replication_factor": 3,
}
CONSUMER_GROUP_ID = f"consumer_group_{DAG_ID}_{ENV_ID}".replace("_", "-")
CONNECTION_ID = f"connection_{DAG_ID}_{ENV_ID}"
PORT = "9092"
BOOTSTRAP_URL = f"bootstrap.{CLUSTER_ID}.{LOCATION}.managedkafka.{PROJECT_ID}.cloud.goog:{PORT}"

log = logging.getLogger(__name__)


def producer():
    """Produce and submit 10 messages"""

    for i in range(10):
        now = datetime.now()
        datetime_string = now.strftime("%Y-%m-%d %H:%M:%S")

        message_data = {"random_id": f"{ENV_ID}_{random.randint(1, 100)}", "date_time": datetime_string}

        yield (
            json.dumps(i),
            json.dumps(message_data),
        )


def consumer(message):
    "Take in consumed messages and print its contents to the logs."

    message_content = json.loads(message.value())
    random_id = message_content["random_id"]
    date_time = message_content["date_time"]
    log.info("id: %s, date_time: %s", random_id, date_time)


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "managed_kafka", "consumer_group"],
) as dag:
    create_cluster = ManagedKafkaCreateClusterOperator(
        task_id="create_cluster",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster=CLUSTER_CONF,
        cluster_id=CLUSTER_ID,
    )

    create_topic = ManagedKafkaCreateTopicOperator(
        task_id="create_topic",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
        topic_id=TOPIC_ID,
        topic=TOPIC_CONF,
    )

    @task
    def create_connection(connection_id: str):
        conn_extra = {
            "bootstrap.servers": BOOTSTRAP_URL,
            "security.protocol": "SASL_SSL",
            "sasl.mechanisms": "OAUTHBEARER",
            "group.id": CONSUMER_GROUP_ID,
        }
        conn_extra_json = json.dumps(conn_extra)
        connection: dict[str, Any] = {"conn_type": "kafka", "extra": conn_extra_json}
        create_airflow_connection(
            connection_id=connection_id,
            connection_conf=connection,
        )

    create_connection_task = create_connection(connection_id=CONNECTION_ID)

    # [START how_to_cloud_managed_kafka_produce_to_topic_operator]
    produce_to_topic = ProduceToTopicOperator(
        task_id="produce_to_topic",
        kafka_config_id=CONNECTION_ID,
        topic=TOPIC_ID,
        producer_function=producer,
        poll_timeout=10,
    )
    # [END how_to_cloud_managed_kafka_produce_to_topic_operator]

    # [START how_to_cloud_managed_kafka_consume_from_topic_operator]
    consume_from_topic = ConsumeFromTopicOperator(
        task_id="consume_from_topic",
        kafka_config_id=CONNECTION_ID,
        topics=[TOPIC_ID],
        apply_function=consumer,
        poll_timeout=20,
        max_messages=20,
        max_batch_size=20,
    )
    # [END how_to_cloud_managed_kafka_consume_from_topic_operator]

    # [START how_to_cloud_managed_kafka_update_consumer_group_operator]
    update_consumer_group = ManagedKafkaUpdateConsumerGroupOperator(
        task_id="update_consumer_group",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
        consumer_group_id=CONSUMER_GROUP_ID,
        consumer_group={
            "topics": {},
        },
        update_mask={"paths": ["topics"]},
    )
    # [END how_to_cloud_managed_kafka_update_consumer_group_operator]

    # [START how_to_cloud_managed_kafka_get_consumer_group_operator]
    get_consumer_group = ManagedKafkaGetConsumerGroupOperator(
        task_id="get_consumer_group",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
        consumer_group_id=CONSUMER_GROUP_ID,
    )
    # [END how_to_cloud_managed_kafka_get_consumer_group_operator]

    # [START how_to_cloud_managed_kafka_delete_consumer_group_operator]
    delete_consumer_group = ManagedKafkaDeleteConsumerGroupOperator(
        task_id="delete_consumer_group",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
        consumer_group_id=CONSUMER_GROUP_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END how_to_cloud_managed_kafka_delete_consumer_group_operator]

    delete_cluster = ManagedKafkaDeleteClusterOperator(
        task_id="delete_cluster",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    # [START how_to_cloud_managed_kafka_list_consumer_group_operator]
    list_consumer_groups = ManagedKafkaListConsumerGroupsOperator(
        task_id="list_consumer_groups",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=CLUSTER_ID,
    )
    # [END how_to_cloud_managed_kafka_list_consumer_group_operator]

    (
        # TEST SETUP
        create_cluster
        >> create_topic
        >> create_connection_task
        >> produce_to_topic
        >> consume_from_topic
        # TEST BODY
        >> update_consumer_group
        >> get_consumer_group
        >> list_consumer_groups
        >> delete_consumer_group
        # TEST TEARDOWN
        >> delete_cluster
    )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.translate_v3.types import (
    BatchTranslateDocumentResponse,
    Dataset,
    TranslateDocumentResponse,
    automl_translation,
    translation_service,
)

from airflow.providers.google.cloud.hooks.translate import TranslateHook
from airflow.providers.google.cloud.operators.translate import (
    CloudTranslateTextOperator,
    TranslateCreateDatasetOperator,
    TranslateCreateGlossaryOperator,
    TranslateCreateModelOperator,
    TranslateDatasetsListOperator,
    TranslateDeleteDatasetOperator,
    TranslateDeleteGlossaryOperator,
    TranslateDeleteModelOperator,
    TranslateDocumentBatchOperator,
    TranslateDocumentOperator,
    TranslateImportDataOperator,
    TranslateListGlossariesOperator,
    TranslateModelsListOperator,
    TranslateTextBatchOperator,
    TranslateTextOperator,
    TranslateUpdateGlossaryOperator,
)

GCP_CONN_ID = "google_cloud_default"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
PROJECT_ID = "test-project-id"
DATASET_ID = "sample_ds_id"
MODEL_ID = "sample_model_id"
GLOSSARY_ID = "sample_glossary_id"
TIMEOUT_VALUE = 30
LOCATION = "location_id"


class TestCloudTranslate:
    @mock.patch("airflow.providers.google.cloud.operators.translate.CloudTranslateHook")
    def test_minimal_green_path(self, mock_hook):
        mock_hook.return_value.translate.return_value = [
            {
                "translatedText": "Yellowing self Gęśle",
                "detectedSourceLanguage": "pl",
                "model": "base",
                "input": "zażółć gęślą jaźń",
            }
        ]
        op = CloudTranslateTextOperator(
            values=["zażółć gęślą jaźń"],
            target_language="en",
            format_="text",
            source_language=None,
            model="base",
            gcp_conn_id=GCP_CONN_ID,
            task_id="id",
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        return_value = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.translate.assert_called_once_with(
            values=["zażółć gęślą jaźń"],
            target_language="en",
            format_="text",
            source_language=None,
            model="base",
        )
        assert return_value == [
            {
                "translatedText": "Yellowing self Gęśle",
                "detectedSourceLanguage": "pl",
                "model": "base",
                "input": "zażółć gęślą jaźń",
            }
        ]


class TestTranslateText:
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook):
        translation_result_data = {
            "translations": [
                {"translated_text": "Hello World!", "model": "", "detected_language_code": ""},
                {
                    "translated_text": "Can you get me a cup of coffee, please?",
                    "model": "",
                    "detected_language_code": "",
                },
            ],
            "glossary_translations": [],
        }
        mock_hook.return_value.translate_text.return_value = translation_result_data
        data_to_translate = ["Ciao mondo!", "Mi puoi prendere una tazza di caffè, per favore?"]
        op = TranslateTextOperator(
            task_id="task_id",
            contents=data_to_translate,
            source_language_code="it",
            target_language_code="en",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=None,
            model=None,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.translate_text.assert_called_once_with(
            contents=data_to_translate,
            source_language_code="it",
            target_language_code="en",
            mime_type=None,
            location=None,
            labels=None,
            model=None,
            transliteration_config=None,
            glossary_config=None,
            timeout=TIMEOUT_VALUE,
            retry=None,
            metadata=(),
        )
        assert translation_result_data == result


class TestTranslateTextBatchOperator:
    @mock.patch("airflow.providers.google.cloud.links.translate.TranslateTextBatchLink.persist")
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook, mock_link_persist):
        input_config_item = {
            "gcs_source": {"input_uri": "gs://source_bucket_uri/sample_data_src_lang.txt"},
            "mime_type": "text/plain",
        }
        SRC_LANG_CODE = "src_lang_code"
        TARGET_LANG_CODES = ["target_lang_code1", "target_lang_code2"]
        TIMEOUT = 30
        INPUT_CONFIGS = [input_config_item]
        OUTPUT_CONFIG = {"gcs_destination": {"output_uri_prefix": "gs://source_bucket_uri/output/"}}
        batch_translation_results_data = {"batch_text_translate_results": OUTPUT_CONFIG["gcs_destination"]}
        mock_hook.return_value.batch_translate_text.return_value = batch_translation_results_data

        op = TranslateTextBatchOperator(
            task_id="task_id_test",
            project_id=PROJECT_ID,
            source_language_code=SRC_LANG_CODE,
            target_language_codes=TARGET_LANG_CODES,
            location=LOCATION,
            models=None,
            glossaries=None,
            input_configs=INPUT_CONFIGS,
            output_config=OUTPUT_CONFIG,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT,
            retry=None,
        )
        context = {"ti": mock.MagicMock()}
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.batch_translate_text.assert_called_once_with(
            project_id=PROJECT_ID,
            source_language_code=SRC_LANG_CODE,
            target_language_codes=TARGET_LANG_CODES,
            location=LOCATION,
            input_configs=INPUT_CONFIGS,
            output_config=OUTPUT_CONFIG,
            timeout=TIMEOUT,
            models=None,
            glossaries=None,
            labels=None,
            retry=None,
            metadata=(),
        )
        assert batch_translation_results_data == result

        mock_link_persist.assert_called_once_with(
            context=context,
            project_id=PROJECT_ID,
            output_config=OUTPUT_CONFIG,
        )


class TestTranslateDatasetCreate:
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslationNativeDatasetLink.persist")
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook, mock_link_persist):
        DS_CREATION_RESULT_SAMPLE = {
            "display_name": "",
            "example_count": 0,
            "name": f"projects/{PROJECT_ID}/locations/{LOCATION}/datasets/{DATASET_ID}",
            "source_language_code": "",
            "target_language_code": "",
            "test_example_count": 0,
            "train_example_count": 0,
            "validate_example_count": 0,
        }
        sample_operation = mock.MagicMock()
        sample_operation.result.return_value = automl_translation.Dataset(DS_CREATION_RESULT_SAMPLE)

        mock_hook.return_value.create_dataset.return_value = sample_operation
        mock_hook.return_value.wait_for_operation_result.side_effect = lambda operation: operation.result()
        mock_hook.return_value.extract_object_id = TranslateHook.extract_object_id

        DATASET_DATA = {
            "display_name": "sample ds name",
            "source_language_code": "es",
            "target_language_code": "uk",
        }
        op = TranslateCreateDatasetOperator(
            task_id="task_id",
            dataset=DATASET_DATA,
            project_id=PROJECT_ID,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=None,
        )
        mock_ti = mock.MagicMock()
        mock_context = {"ti": mock_ti}
        result = op.execute(context=mock_context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_dataset.assert_called_once_with(
            dataset=DATASET_DATA,
            project_id=PROJECT_ID,
            location=LOCATION,
            timeout=TIMEOUT_VALUE,
            retry=None,
            metadata=(),
        )
        mock_ti.xcom_push.assert_any_call(key="dataset_id", value=DATASET_ID)
        mock_link_persist.assert_called_once_with(
            context=mock_context,
            dataset_id=DATASET_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        assert result == DS_CREATION_RESULT_SAMPLE


class TestTranslateListDatasets:
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslationDatasetsListLink.persist")
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook, mock_link_persist):
        DS_ID_1 = "sample_ds_1"
        DS_ID_2 = "sample_ds_2"
        dataset_result_1 = automl_translation.Dataset(
            dict(
                name=f"projects/{PROJECT_ID}/locations/{LOCATION}/datasets/{DS_ID_1}",
                display_name="ds1_display_name",
            )
        )
        dataset_result_2 = automl_translation.Dataset(
            dict(
                name=f"projects/{PROJECT_ID}/locations/{LOCATION}/datasets/{DS_ID_2}",
                display_name="ds1_display_name",
            )
        )
        mock_hook.return_value.list_datasets.return_value = [dataset_result_1, dataset_result_2]
        mock_hook.return_value.extract_object_id = TranslateHook.extract_object_id

        op = TranslateDatasetsListOperator(
            task_id="task_id",
            project_id=PROJECT_ID,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.list_datasets.assert_called_once_with(
            project_id=PROJECT_ID,
            location=LOCATION,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
            metadata=(),
        )
        mock_link_persist.assert_called_once_with(
            context=context,
            project_id=PROJECT_ID,
        )
        assert result == [DS_ID_1, DS_ID_2]


class TestTranslateImportData:
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslationNativeDatasetLink.persist")
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook, mock_link_persist):
        INPUT_CONFIG = {
            "input_files": [{"usage": "UNASSIGNED", "gcs_source": {"input_uri": "import data gcs path"}}]
        }
        mock_hook.return_value.import_dataset_data.return_value = mock.MagicMock()

        SAMPLE_DATASET = {
            "name": "sample_translation_dataset",
            "example_count": None,
            "source_language_code": "en",
            "target_language_code": "es",
        }
        INITIAL_DS_SIZE = 1
        FINAL_DS_SIZE = 101
        INITIAL_DS = {**SAMPLE_DATASET, "example_count": INITIAL_DS_SIZE}
        FINAL_DS = {**SAMPLE_DATASET, "example_count": FINAL_DS_SIZE}

        mock_hook.return_value.get_dataset.side_effect = [Dataset(INITIAL_DS), Dataset(FINAL_DS)]
        op = TranslateImportDataOperator(
            task_id="task_id",
            dataset_id=DATASET_ID,
            input_config=INPUT_CONFIG,
            project_id=PROJECT_ID,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
        )
        context = mock.MagicMock()
        res = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.import_dataset_data.assert_called_once_with(
            dataset_id=DATASET_ID,
            input_config=INPUT_CONFIG,
            project_id=PROJECT_ID,
            location=LOCATION,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
            metadata=(),
        )
        mock_link_persist.assert_called_once_with(
            context=context,
            dataset_id=DATASET_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        assert res["total_imported"] == FINAL_DS_SIZE - INITIAL_DS_SIZE


class TestTranslateDeleteData:
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook):
        m_delete_method_result = mock.MagicMock()
        mock_hook.return_value.delete_dataset.return_value = m_delete_method_result

        wait_for_done = mock_hook.return_value.wait_for_operation_done

        op = TranslateDeleteDatasetOperator(
            task_id="task_id",
            dataset_id=DATASET_ID,
            project_id=PROJECT_ID,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_dataset.assert_called_once_with(
            dataset_id=DATASET_ID,
            project_id=PROJECT_ID,
            location=LOCATION,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
            metadata=(),
        )
        wait_for_done.assert_called_once_with(operation=m_delete_method_result, timeout=TIMEOUT_VALUE)


class TestTranslateModelCreate:
    @mock.patch("airflow.providers.google.cloud.links.translate.TranslationModelLink.persist")
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook, mock_link_persist):
        MODEL_DISPLAY_NAME = "model_display_name_01"
        MODEL_CREATION_RESULT_SAMPLE = {
            "display_name": MODEL_DISPLAY_NAME,
            "name": f"projects/{PROJECT_ID}/locations/{LOCATION}/models/{MODEL_ID}",
            "dataset": f"projects/{PROJECT_ID}/locations/{LOCATION}/datasets/{DATASET_ID}",
            "source_language_code": "",
            "target_language_code": "",
            "create_time": "2024-11-15T14:05:00Z",
            "update_time": "2024-11-16T01:09:03Z",
            "test_example_count": 1000,
            "train_example_count": 115,
            "validate_example_count": 140,
        }
        sample_operation = mock.MagicMock()
        sample_operation.result.return_value = automl_translation.Model(MODEL_CREATION_RESULT_SAMPLE)

        mock_hook.return_value.create_model.return_value = sample_operation
        mock_hook.return_value.wait_for_operation_result.side_effect = lambda operation: operation.result()
        mock_hook.return_value.extract_object_id = TranslateHook.extract_object_id
        op = TranslateCreateModelOperator(
            task_id="task_id",
            display_name=MODEL_DISPLAY_NAME,
            dataset_id=DATASET_ID,
            project_id=PROJECT_ID,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=None,
        )
        mock_ti = mock.MagicMock()
        mock_context = {"ti": mock_ti}
        result = op.execute(context=mock_context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_model.assert_called_once_with(
            display_name=MODEL_DISPLAY_NAME,
            dataset_id=DATASET_ID,
            project_id=PROJECT_ID,
            location=LOCATION,
            timeout=TIMEOUT_VALUE,
            retry=None,
            metadata=(),
        )
        mock_ti.xcom_push.assert_any_call(key="model_id", value=MODEL_ID)
        mock_link_persist.assert_called_once_with(
            context=mock_context,
            model_id=MODEL_ID,
            project_id=PROJECT_ID,
            dataset_id=DATASET_ID,
            location=LOCATION,
        )
        assert result == MODEL_CREATION_RESULT_SAMPLE


class TestTranslateListModels:
    @mock.patch("airflow.providers.google.cloud.links.translate.TranslationModelsListLink.persist")
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook, mock_link_persist):
        MODEL_ID_1 = "sample_model_1"
        MODEL_ID_2 = "sample_model_2"
        model_result_1 = automl_translation.Model(
            dict(
                display_name="model_1_display_name",
                name=f"projects/{PROJECT_ID}/locations/{LOCATION}/models/{MODEL_ID_1}",
                dataset=f"projects/{PROJECT_ID}/locations/{LOCATION}/datasets/ds_for_model_1",
                source_language_code="en",
                target_language_code="es",
            )
        )
        model_result_2 = automl_translation.Model(
            dict(
                display_name="model_2_display_name",
                name=f"projects/{PROJECT_ID}/locations/{LOCATION}/models/{MODEL_ID_2}",
                dataset=f"projects/{PROJECT_ID}/locations/{LOCATION}/datasets/ds_for_model_2",
                source_language_code="uk",
                target_language_code="en",
            )
        )
        mock_hook.return_value.list_models.return_value = [model_result_1, model_result_2]
        mock_hook.return_value.extract_object_id = TranslateHook.extract_object_id

        op = TranslateModelsListOperator(
            task_id="task_id",
            project_id=PROJECT_ID,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.list_models.assert_called_once_with(
            project_id=PROJECT_ID,
            location=LOCATION,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
            metadata=(),
        )
        assert result == [MODEL_ID_1, MODEL_ID_2]
        mock_link_persist.assert_called_once_with(
            context=context,
            project_id=PROJECT_ID,
        )


class TestTranslateDeleteModel:
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook):
        m_delete_method_result = mock.MagicMock()
        mock_hook.return_value.delete_model.return_value = m_delete_method_result
        wait_for_done = mock_hook.return_value.wait_for_operation_done

        op = TranslateDeleteModelOperator(
            task_id="task_id",
            model_id=MODEL_ID,
            project_id=PROJECT_ID,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_model.assert_called_once_with(
            model_id=MODEL_ID,
            project_id=PROJECT_ID,
            location=LOCATION,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
            metadata=(),
        )
        wait_for_done.assert_called_once_with(operation=m_delete_method_result, timeout=TIMEOUT_VALUE)


class TestTranslateDocumentBatchOperator:
    @mock.patch("airflow.providers.google.cloud.links.translate.TranslateResultByOutputConfigLink.persist")
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook, mock_link_persist):
        input_config_item_1 = {
            "gcs_source": {"input_uri": "gs://source_bucket_uri/sample_data_src_lang_1.txt"},
        }
        input_config_item_2 = {
            "gcs_source": {"input_uri": "gs://source_bucket_uri/sample_data_src_lang_2.txt"},
        }
        SRC_LANG_CODE = "src_lang_code"
        TARGET_LANG_CODES = ["target_lang_code1", "target_lang_code2"]
        TIMEOUT = 30
        INPUT_CONFIGS = [input_config_item_1, input_config_item_2]
        OUTPUT_CONFIG = {"gcs_destination": {"output_uri_prefix": "gs://source_bucket_uri/output/"}}
        BATCH_DOC_TRANSLATION_RESULT = {
            "submit_time": "2024-12-01T00:01:16Z",
            "end_time": "2024-12-01T00:10:01Z",
            "failed_characters": "0",
            "failed_pages": "0",
            "total_billable_characters": "0",
            "total_billable_pages": "6",
            "total_characters": "4240",
            "total_pages": "6",
            "translated_characters": "4240",
            "translated_pages": "6",
        }
        sample_operation = mock.MagicMock()
        sample_operation.result.return_value = BatchTranslateDocumentResponse(BATCH_DOC_TRANSLATION_RESULT)

        mock_hook.return_value.batch_translate_document.return_value = sample_operation
        mock_hook.return_value.wait_for_operation_result.side_effect = lambda operation: operation.result()

        op = TranslateDocumentBatchOperator(
            task_id="task_id_test",
            project_id=PROJECT_ID,
            source_language_code=SRC_LANG_CODE,
            target_language_codes=TARGET_LANG_CODES,
            location=LOCATION,
            models=None,
            glossaries=None,
            input_configs=INPUT_CONFIGS,
            output_config=OUTPUT_CONFIG,
            customized_attribution=None,
            format_conversions=None,
            enable_shadow_removal_native_pdf=False,
            enable_rotation_correction=False,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            metadata=(),
            timeout=TIMEOUT,
            retry=None,
        )
        context = {"ti": mock.MagicMock()}
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.batch_translate_document.assert_called_once_with(
            project_id=PROJECT_ID,
            source_language_code=SRC_LANG_CODE,
            target_language_codes=TARGET_LANG_CODES,
            location=LOCATION,
            input_configs=INPUT_CONFIGS,
            output_config=OUTPUT_CONFIG,
            customized_attribution=None,
            format_conversions=None,
            enable_shadow_removal_native_pdf=False,
            enable_rotation_correction=False,
            timeout=TIMEOUT,
            models=None,
            glossaries=None,
            retry=None,
            metadata=(),
        )

        assert result == BATCH_DOC_TRANSLATION_RESULT
        mock_link_persist.assert_called_once_with(
            context=context,
            project_id=PROJECT_ID,
            output_config=OUTPUT_CONFIG,
        )


class TestTranslateDocumentOperator:
    @mock.patch("airflow.providers.google.cloud.links.translate.TranslateResultByOutputConfigLink.persist")
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook, mock_link_persist):
        SRC_LANG_CODE = "src_lang_code"
        TARGET_LANG_CODE = "target_lang_code1"
        TIMEOUT = 30
        INPUT_CONFIG = {"gcs_source": {"input_uri": "gs://source_bucket_uri/sample_data_src_lang_1.txt"}}
        OUTPUT_CONFIG = {"gcs_destination": {"output_uri_prefix": "gs://source_bucket_uri/output/"}}
        DOC_TRANSLATION_RESULT = {
            "document_translation": {
                "byte_stream_outputs": ["c29tZV9kYXRh"],
                "detected_language_code": "",
                "mime_type": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            },
            "model": f"projects/{PROJECT_ID}/locations/us-central1/models/general/nmt",
        }

        mock_hook.return_value.translate_document.return_value = TranslateDocumentResponse(
            DOC_TRANSLATION_RESULT
        )

        op = TranslateDocumentOperator(
            task_id="task_id_test",
            project_id=PROJECT_ID,
            source_language_code=SRC_LANG_CODE,
            target_language_code=TARGET_LANG_CODE,
            location=LOCATION,
            model=None,
            glossary_config=None,
            labels=None,
            document_input_config=INPUT_CONFIG,
            document_output_config=OUTPUT_CONFIG,
            customized_attribution=None,
            is_translate_native_pdf_only=False,
            enable_shadow_removal_native_pdf=False,
            enable_rotation_correction=False,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT,
            retry=None,
        )
        context = {"ti": mock.MagicMock()}
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.translate_document.assert_called_once_with(
            source_language_code=SRC_LANG_CODE,
            target_language_code=TARGET_LANG_CODE,
            location=LOCATION,
            model=None,
            glossary_config=None,
            labels=None,
            document_input_config=INPUT_CONFIG,
            document_output_config=OUTPUT_CONFIG,
            customized_attribution=None,
            is_translate_native_pdf_only=False,
            enable_shadow_removal_native_pdf=False,
            enable_rotation_correction=False,
            timeout=TIMEOUT,
            retry=None,
            metadata=(),
        )

        assert result == DOC_TRANSLATION_RESULT
        mock_link_persist.assert_called_once_with(
            context=context,
            project_id=PROJECT_ID,
            output_config=OUTPUT_CONFIG,
        )


class TestTranslateGlossaryCreate:
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook):
        GLOSSARY_CREATION_RESULT = {
            "name": f"projects/{PROJECT_ID}/locations/{LOCATION}/glossaries/{GLOSSARY_ID}",
            "display_name": f"{GLOSSARY_ID}",
            "entry_count": 42,
            "input_config": {"gcs_source": {"input_uri": "gs://input_bucket_path/glossary.csv"}},
            "language_pair": {"source_language_code": "en", "target_language_code": "es"},
            "submit_time": "2024-11-17T14:05:00Z",
            "end_time": "2024-11-17T17:09:03Z",
        }
        sample_operation = mock.MagicMock()
        sample_operation.result.return_value = translation_service.Glossary(GLOSSARY_CREATION_RESULT)

        mock_hook.return_value.create_glossary.return_value = sample_operation
        mock_hook.return_value.wait_for_operation_result.side_effect = lambda operation: operation.result()
        mock_hook.return_value.extract_object_id = TranslateHook.extract_object_id

        GLOSSARY_FILE_INPUT = {"gcs_source": {"input_uri": "gs://RESOURCE_BUCKET/glossary_sample.tsv"}}
        op = TranslateCreateGlossaryOperator(
            task_id="task_id",
            glossary_id=f"{GLOSSARY_ID}",
            input_config=GLOSSARY_FILE_INPUT,
            language_pair={"source_language_code": "en", "target_language_code": "es"},
            language_codes_set=None,
            project_id=PROJECT_ID,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=None,
        )
        mock_ti = mock.MagicMock()
        mock_context = {"ti": mock_ti}
        result = op.execute(context=mock_context)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_glossary.assert_called_once_with(
            glossary_id=f"{GLOSSARY_ID}",
            input_config=GLOSSARY_FILE_INPUT,
            language_pair={"source_language_code": "en", "target_language_code": "es"},
            language_codes_set=None,
            project_id=PROJECT_ID,
            location=LOCATION,
            timeout=TIMEOUT_VALUE,
            retry=None,
            metadata=(),
        )
        mock_ti.xcom_push.assert_any_call(key="glossary_id", value=GLOSSARY_ID)
        assert result == GLOSSARY_CREATION_RESULT


class TestTranslateGlossaryUpdate:
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook):
        UPDATE_GLOSSARY_RESULT = {
            "name": f"projects/{PROJECT_ID}/locations/{LOCATION}/glossaries/{GLOSSARY_ID}",
            "display_name": "new_glossary_display_name",
            "entry_count": 42,
            "input_config": {"gcs_source": {"input_uri": "gs://input_bucket_path/glossary_updated.csv"}},
            "language_pair": {"source_language_code": "en", "target_language_code": "es"},
            "submit_time": "2024-11-17T14:05:00Z",
            "end_time": "2024-11-17T17:09:03Z",
        }
        get_glossary_mock = mock.MagicMock()
        mock_hook.return_value.get_glossary.return_value = get_glossary_mock
        sample_operation = mock.MagicMock()
        sample_operation.result.return_value = translation_service.Glossary(UPDATE_GLOSSARY_RESULT)

        mock_hook.return_value.update_glossary.return_value = sample_operation
        mock_hook.return_value.wait_for_operation_result.side_effect = lambda operation: operation.result()

        UPDATE_GLOSSARY_FILE_INPUT = {"gcs_source": {"input_uri": "gs://RESOURCE_BUCKET/glossary_sample.tsv"}}
        op = TranslateUpdateGlossaryOperator(
            task_id="task_id",
            new_input_config=UPDATE_GLOSSARY_FILE_INPUT,
            new_display_name="new_glossary_display_name",
            glossary_id=GLOSSARY_ID,
            project_id=PROJECT_ID,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=None,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_glossary.assert_called_once_with(
            glossary=get_glossary_mock,
            new_input_config=UPDATE_GLOSSARY_FILE_INPUT,
            new_display_name="new_glossary_display_name",
            timeout=TIMEOUT_VALUE,
            retry=None,
            metadata=(),
        )
        assert result == UPDATE_GLOSSARY_RESULT


class TestTranslateListGlossaries:
    @mock.patch("airflow.providers.google.cloud.links.translate.TranslationGlossariesListLink.persist")
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook, mock_link_persist):
        GLOSSARY_ID_1 = "sample_glossary_1"
        GLOSSARY_ID_2 = "sample_glossary_2"
        glossary_result_1 = translation_service.Glossary(
            dict(
                name=f"projects/{PROJECT_ID}/locations/{LOCATION}/glossaries/{GLOSSARY_ID_1}",
                display_name=f"{GLOSSARY_ID_1}",
                entry_count=100,
                input_config={"gcs_source": {"input_uri": "gs://input1.csv"}},
                language_pair={"source_language_code": "en", "target_language_code": "es"},
                submit_time="2024-11-17T14:05:00Z",
                end_time="2024-11-17T17:09:03Z",
            )
        )
        glossary_result_2 = translation_service.Glossary(
            dict(
                name=f"projects/{PROJECT_ID}/locations/{LOCATION}/glossaries/{GLOSSARY_ID_2}",
                display_name=f"{GLOSSARY_ID_2}",
                entry_count=200,
                input_config={"gcs_source": {"input_uri": "gs://input2.csv"}},
                language_pair={"source_language_code": "es", "target_language_code": "en"},
                submit_time="2024-11-17T14:05:00Z",
                end_time="2024-11-17T17:09:03Z",
            )
        )
        mock_hook.return_value.list_glossaries.return_value = [glossary_result_1, glossary_result_2]
        mock_hook.return_value.extract_object_id = TranslateHook.extract_object_id

        op = TranslateListGlossariesOperator(
            task_id="task_id",
            project_id=PROJECT_ID,
            page_size=100,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.list_glossaries.assert_called_once_with(
            project_id=PROJECT_ID,
            page_size=100,
            filter_str=None,
            page_token=None,
            location=LOCATION,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
            metadata=(),
        )
        assert result == [GLOSSARY_ID_1, GLOSSARY_ID_2]
        mock_link_persist.assert_called_once_with(
            context=context,
            project_id=PROJECT_ID,
        )


class TestTranslateDeleteGlossary:
    @mock.patch("airflow.providers.google.cloud.operators.translate.TranslateHook")
    def test_minimal_green_path(self, mock_hook):
        DELETION_RESULT_SAMPLE = {
            "submit_time": "2024-11-17T14:05:00Z",
            "end_time": "2024-11-17T17:09:03Z",
            "name": f"projects/{PROJECT_ID}/locations/{LOCATION}/glossaries/{GLOSSARY_ID}",
        }
        sample_operation = mock.MagicMock()
        sample_operation.result.return_value = translation_service.DeleteGlossaryResponse(
            DELETION_RESULT_SAMPLE
        )
        gl_delete_method = mock_hook.return_value.delete_glossary
        gl_delete_method.return_value = sample_operation

        mock_hook.return_value.wait_for_operation_result.side_effect = lambda operation: operation.result()

        op = TranslateDeleteGlossaryOperator(
            task_id="task_id",
            glossary_id=GLOSSARY_ID,
            project_id=PROJECT_ID,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        gl_delete_method.assert_called_once_with(
            glossary_id=GLOSSARY_ID,
            project_id=PROJECT_ID,
            location=LOCATION,
            timeout=TIMEOUT_VALUE,
            retry=DEFAULT,
            metadata=(),
        )
        assert result == DELETION_RESULT_SAMPLE



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.operators.vision import (
    CloudVisionAddProductToProductSetOperator,
    CloudVisionCreateProductOperator,
    CloudVisionCreateProductSetOperator,
    CloudVisionCreateReferenceImageOperator,
    CloudVisionDeleteProductOperator,
    CloudVisionDeleteProductSetOperator,
    CloudVisionDeleteReferenceImageOperator,
    CloudVisionGetProductOperator,
    CloudVisionGetProductSetOperator,
    CloudVisionRemoveProductFromProductSetOperator,
    CloudVisionUpdateProductOperator,
    CloudVisionUpdateProductSetOperator,
)
from airflow.providers.google.cloud.transfers.gcs_to_gcs import GCSToGCSOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

# [START howto_operator_vision_retry_import]
from google.api_core.retry import Retry

# [END howto_operator_vision_retry_import]
# [START howto_operator_vision_product_set_import]
from google.cloud.vision_v1.types import ProductSet

# [END howto_operator_vision_product_set_import]
# [START howto_operator_vision_product_import]
from google.cloud.vision_v1.types import Product  # isort:skip

# [END howto_operator_vision_product_import]
# [START howto_operator_vision_reference_image_import]
from google.cloud.vision_v1.types import ReferenceImage

# [END howto_operator_vision_reference_image_import]
# [START howto_operator_vision_enums_import]
from google.cloud.vision_v1 import Feature  # isort:skip
from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

# [END howto_operator_vision_enums_import]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "gcp_vision_autogenerated_id"

LOCATION = "europe-west1"

BUCKET_NAME = f"bucket-{DAG_ID}-{ENV_ID}".replace("_", "-")
FILE_NAME = "image1.jpg"

GCP_VISION_PRODUCT_SET_ID = "product_set_explicit_id"
GCP_VISION_PRODUCT_ID = "product_explicit_id"
GCP_VISION_REFERENCE_IMAGE_ID = f"reference_image_explicit_id-{ENV_ID}"

VISION_IMAGE_URL = f"gs://{BUCKET_NAME}/{FILE_NAME}"

# [START howto_operator_vision_product_set]
product_set = ProductSet(display_name="My Product Set")
# [END howto_operator_vision_product_set]

# [START howto_operator_vision_product]
product = Product(display_name="My Product 1", product_category="toys")
# [END howto_operator_vision_product]

# [START howto_operator_vision_reference_image]
reference_image = ReferenceImage(uri=VISION_IMAGE_URL)
# [END howto_operator_vision_reference_image]

# [START howto_operator_vision_annotate_image_request]
annotate_image_request = {
    "image": {"source": {"image_uri": VISION_IMAGE_URL}},
    "features": [{"type_": Feature.Type.LOGO_DETECTION}],
}
# [END howto_operator_vision_annotate_image_request]

# [START howto_operator_vision_detect_image_param]
DETECT_IMAGE = {"source": {"image_uri": VISION_IMAGE_URL}}
# [END howto_operator_vision_detect_image_param]

# Public bucket holding the sample data
BUCKET_NAME_SRC = "cloud-samples-data"
# Path to the data inside the public bucket
PATH_SRC = "vision/ocr/sign.jpg"

with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "vision", "autogenerated"],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket", project_id=PROJECT_ID, bucket_name=BUCKET_NAME
    )

    copy_single_file = GCSToGCSOperator(
        task_id="copy_single_gcs_file",
        source_bucket=BUCKET_NAME_SRC,
        source_object=PATH_SRC,
        destination_bucket=BUCKET_NAME,
        destination_object=FILE_NAME,
    )

    # [START howto_operator_vision_product_set_create]
    product_set_create = CloudVisionCreateProductSetOperator(
        location=LOCATION,
        product_set=product_set,
        retry=Retry(maximum=10.0),
        timeout=5,
        task_id="product_set_create",
    )
    # [END howto_operator_vision_product_set_create]

    product_set_create_output = "{{ task_instance.xcom_pull('product_set_create') }}"

    # [START howto_operator_vision_product_set_get]
    product_set_get = CloudVisionGetProductSetOperator(
        location=LOCATION,
        product_set_id=product_set_create_output,
        task_id="product_set_get",
    )
    # [END howto_operator_vision_product_set_get]

    # [START howto_operator_vision_product_set_update]
    product_set_update = CloudVisionUpdateProductSetOperator(
        location=LOCATION,
        product_set_id=product_set_create_output,
        product_set=ProductSet(display_name="My Product Set 2"),
        task_id="product_set_update",
    )
    # [END howto_operator_vision_product_set_update]

    # [START howto_operator_vision_product_set_delete]
    product_set_delete = CloudVisionDeleteProductSetOperator(
        location=LOCATION,
        product_set_id=product_set_create_output,
        task_id="product_set_delete",
    )
    # [END howto_operator_vision_product_set_delete]

    # [START howto_operator_vision_product_create]
    product_create = CloudVisionCreateProductOperator(
        location=LOCATION,
        product=product,
        retry=Retry(maximum=10.0),
        timeout=5,
        task_id="product_create",
    )
    # [END howto_operator_vision_product_create]

    # [START howto_operator_vision_product_get]
    product_get = CloudVisionGetProductOperator(
        location=LOCATION,
        product_id="{{ task_instance.xcom_pull('product_create') }}",
        task_id="product_get",
    )
    # [END howto_operator_vision_product_get]

    # [START howto_operator_vision_product_update]
    product_update = CloudVisionUpdateProductOperator(
        location=LOCATION,
        product_id="{{ task_instance.xcom_pull('product_create') }}",
        product=Product(display_name="My Product 2", description="My updated description"),
        task_id="product_update",
    )
    # [END howto_operator_vision_product_update]

    # [START howto_operator_vision_product_delete]
    product_delete = CloudVisionDeleteProductOperator(
        location=LOCATION,
        product_id="{{ task_instance.xcom_pull('product_create') }}",
        task_id="product_delete",
    )
    # [END howto_operator_vision_product_delete]

    # [START howto_operator_vision_reference_image_create]
    reference_image_create = CloudVisionCreateReferenceImageOperator(
        location=LOCATION,
        reference_image=reference_image,
        product_id="{{ task_instance.xcom_pull('product_create') }}",
        reference_image_id=GCP_VISION_REFERENCE_IMAGE_ID,
        retry=Retry(maximum=10.0),
        timeout=5,
        task_id="reference_image_create",
    )
    # [END howto_operator_vision_reference_image_create]

    # [START howto_operator_vision_reference_image_delete]
    reference_image_delete = CloudVisionDeleteReferenceImageOperator(
        location=LOCATION,
        product_id="{{ task_instance.xcom_pull('product_create') }}",
        reference_image_id=GCP_VISION_REFERENCE_IMAGE_ID,
        retry=Retry(maximum=10.0),
        timeout=5,
        task_id="reference_image_delete",
    )
    # [END howto_operator_vision_reference_image_delete]

    # [START howto_operator_vision_add_product_to_product_set]
    add_product_to_product_set = CloudVisionAddProductToProductSetOperator(
        location=LOCATION,
        product_set_id=product_set_create_output,
        product_id="{{ task_instance.xcom_pull('product_create') }}",
        retry=Retry(maximum=10.0),
        timeout=5,
        task_id="add_product_to_product_set",
    )
    # [END howto_operator_vision_add_product_to_product_set]

    # [START howto_operator_vision_remove_product_from_product_set]
    remove_product_from_product_set = CloudVisionRemoveProductFromProductSetOperator(
        location=LOCATION,
        product_set_id=product_set_create_output,
        product_id="{{ task_instance.xcom_pull('product_create') }}",
        retry=Retry(maximum=10.0),
        timeout=5,
        task_id="remove_product_from_product_set",
    )
    # [END howto_operator_vision_remove_product_from_product_set]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    chain(
        # TEST SETUP
        create_bucket,
        copy_single_file,
        # TEST BODY
        product_create,
        product_get,
        product_update,
        product_set_create,
        product_set_get,
        product_set_update,
        reference_image_create,
        add_product_to_product_set,
        # TEST TEARDOWN
        reference_image_delete,
        remove_product_from_product_set,
        product_set_delete,
        product_delete,
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Cloud Looker operators."""

from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.looker import LookerHook
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

if TYPE_CHECKING:
    from airflow.utils.context import Context


class LookerStartPdtBuildOperator(GoogleCloudBaseOperator):
    """
    Submits a PDT materialization job to Looker.

    :param looker_conn_id: Required. The connection ID to use connecting to Looker.
    :param model: Required. The model of the PDT to start building.
    :param view: Required. The view of the PDT to start building.
    :param query_params: Optional. Additional materialization parameters.
    :param asynchronous: Optional. Flag indicating whether to wait for the job
        to finish or return immediately.
        This is useful for submitting long running jobs and
        waiting on them asynchronously using the LookerCheckPdtBuildSensor
    :param cancel_on_kill: Optional. Flag which indicates whether cancel the
        hook's job or not, when on_kill is called.
    :param wait_time: Optional. Number of seconds between checks for job to be
        ready. Used only if ``asynchronous`` is False.
    :param wait_timeout: Optional. How many seconds wait for job to be ready.
        Used only if ``asynchronous`` is False.
    """

    def __init__(
        self,
        looker_conn_id: str,
        model: str,
        view: str,
        query_params: dict | None = None,
        asynchronous: bool = False,
        cancel_on_kill: bool = True,
        wait_time: int = 10,
        wait_timeout: int | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.model = model
        self.view = view
        self.query_params = query_params
        self.looker_conn_id = looker_conn_id
        self.asynchronous = asynchronous
        self.cancel_on_kill = cancel_on_kill
        self.wait_time = wait_time
        self.wait_timeout = wait_timeout
        self.hook: LookerHook | None = None
        self.materialization_id: str | None = None

    def execute(self, context: Context) -> str:
        self.hook = LookerHook(looker_conn_id=self.looker_conn_id)

        resp = self.hook.start_pdt_build(
            model=self.model,
            view=self.view,
            query_params=self.query_params,
        )

        self.materialization_id = resp.materialization_id

        if not self.materialization_id:
            raise AirflowException(
                f"No `materialization_id` was returned for model: {self.model}, view: {self.view}."
            )

        self.log.info("PDT materialization job submitted successfully. Job id: %s.", self.materialization_id)

        if not self.asynchronous:
            self.hook.wait_for_job(
                materialization_id=self.materialization_id,
                wait_time=self.wait_time,
                timeout=self.wait_timeout,
            )

        return self.materialization_id

    def on_kill(self):
        if self.materialization_id and self.cancel_on_kill:
            self.hook.stop_pdt_build(materialization_id=self.materialization_id)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Hook for Google Cloud Firestore service."""

from __future__ import annotations

import time
from collections.abc import Sequence

from googleapiclient.discovery import build, build_from_document

from airflow.exceptions import AirflowException
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID, GoogleBaseHook

# Time to sleep between active checks of the operation results
TIME_TO_SLEEP_IN_SECONDS = 5


class CloudFirestoreHook(GoogleBaseHook):
    """
    Hook for the Google Firestore APIs.

    All the methods in the hook where project_id is used must be called with
    keyword arguments rather than positional.

    :param api_version: API version used (for example v1 or v1beta1).
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account.
    """

    _conn: build | None = None

    def __init__(
        self,
        api_version: str = "v1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
    ) -> None:
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            impersonation_chain=impersonation_chain,
        )
        self.api_version = api_version

    def get_conn(self):
        """
        Retrieve the connection to Cloud Firestore.

        :return: Google Cloud Firestore services object.
        """
        if not self._conn:
            http_authorized = self._authorize()
            # We cannot use an Authorized Client to retrieve discovery document due to an error in the API.
            # When the authorized customer will send a request to the address below
            # https://www.googleapis.com/discovery/v1/apis/firestore/v1/rest
            # then it will get the message below:
            # > Request contains an invalid argument.
            # At the same time, the Non-Authorized Client has no problems.
            non_authorized_conn = build("firestore", self.api_version, cache_discovery=False)
            self._conn = build_from_document(non_authorized_conn._rootDesc, http=http_authorized)
        return self._conn

    @GoogleBaseHook.fallback_to_default_project_id
    def export_documents(
        self, body: dict, database_id: str = "(default)", project_id: str = PROVIDE_PROJECT_ID
    ) -> None:
        """
        Start a export with the specified configuration.

        :param database_id: The Database ID.
        :param body: The request body.
            See:
            https://firebase.google.com/docs/firestore/reference/rest/v1beta1/projects.databases/exportDocuments
        :param project_id: Optional, Google Cloud Project project_id where the database belongs.
            If set to None or missing, the default project_id from the Google Cloud connection is used.
        """
        service = self.get_conn()

        name = f"projects/{project_id}/databases/{database_id}"

        operation = (
            service.projects()
            .databases()
            .exportDocuments(name=name, body=body)
            .execute(num_retries=self.num_retries)
        )

        self._wait_for_operation_to_complete(operation["name"])

    def _wait_for_operation_to_complete(self, operation_name: str) -> None:
        """
        Wait for the named operation to complete - checks status of the asynchronous call.

        :param operation_name: The name of the operation.
        :return: The response returned by the operation.
        :exception: AirflowException in case error is returned.
        """
        service = self.get_conn()
        while True:
            operation_response = (
                service.projects()
                .databases()
                .operations()
                .get(name=operation_name)
                .execute(num_retries=self.num_retries)
            )
            if operation_response.get("done"):
                response = operation_response.get("response")
                error = operation_response.get("error")
                # Note, according to documentation always either response or error is
                # set when "done" == True
                if error:
                    raise AirflowException(str(error))
                return response
            time.sleep(TIME_TO_SLEEP_IN_SECONDS)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import decimal
from unittest import mock

import pytest

from airflow.models import Connection
from airflow.providers.common.compat.openlineage.facet import (
    OutputDataset,
    SchemaDatasetFacetFields,
)
from airflow.providers.common.sql.hooks.sql import DbApiHook

TASK_ID = "test-mysql-to-gcs"
MYSQL_CONN_ID = "mysql_conn_test"
TZ_QUERY = "SET time_zone = '+00:00'"
SQL = "select 1"
BUCKET = "gs://test"
JSON_FILENAME = "test_{}.ndjson"
CSV_FILENAME = "test_{}.csv"
SCHEMA = [
    {"mode": "REQUIRED", "name": "some_str", "type": "FLOAT"},
    {"mode": "REQUIRED", "name": "some_num", "type": "TIMESTAMP"},
]

ROWS = [("mock_row_content_1", 42), ("mock_row_content_2", 43), ("mock_row_content_3", 44)]
CURSOR_DESCRIPTION = (("some_str", 0, 0, 0, 0, 0, False), ("some_num", 1005, 0, 0, 0, 0, False))
NDJSON_LINES = [
    b'{"some_num": 42, "some_str": "mock_row_content_1"}\n',
    b'{"some_num": 43, "some_str": "mock_row_content_2"}\n',
    b'{"some_num": 44, "some_str": "mock_row_content_3"}\n',
]
CSV_LINES = [
    b"some_str,some_num\r\n",
    b"mock_row_content_1,42\r\n",
    b"mock_row_content_2,43\r\n",
    b"mock_row_content_3,44\r\n",
]
CSV_LINES_PIPE_DELIMITED = [
    b"some_str|some_num\r\n",
    b"mock_row_content_1|42\r\n",
    b"mock_row_content_2|43\r\n",
    b"mock_row_content_3|44\r\n",
]
SCHEMA_FILENAME = "schema_test.json"
SCHEMA_JSON = [
    b'[{"mode": "REQUIRED", "name": "some_str", "type": "FLOAT"}, ',
    b'{"mode": "REQUIRED", "name": "some_num", "type": "STRING"}]',
]
CUSTOM_SCHEMA_JSON = [
    b'[{"mode": "REQUIRED", "name": "some_str", "type": "FLOAT"}, ',
    b'{"mode": "REQUIRED", "name": "some_num", "type": "TIMESTAMP"}]',
]

try:
    from MySQLdb import ProgrammingError

    from airflow.providers.google.cloud.transfers.mysql_to_gcs import MySQLToGCSOperator
except ImportError:
    pytest.skip("MySQL not available", allow_module_level=True)


@pytest.mark.backend("mysql")
class TestMySqlToGoogleCloudStorageOperator:
    def test_init(self):
        """Test MySqlToGoogleCloudStorageOperator instance is properly initialized."""
        op = MySQLToGCSOperator(
            task_id=TASK_ID,
            sql=SQL,
            bucket=BUCKET,
            filename=JSON_FILENAME,
            export_format="CSV",
            field_delimiter="|",
        )
        assert op.task_id == TASK_ID
        assert op.sql == SQL
        assert op.bucket == BUCKET
        assert op.filename == JSON_FILENAME
        assert op.export_format == "csv"
        assert op.field_delimiter == "|"

    @pytest.mark.parametrize(
        "value, schema_type, expected",
        [
            ("string", None, "string"),
            (datetime.date(1970, 1, 2), None, "1970-01-02 00:00:00"),
            (datetime.date(1000, 1, 2), None, "1000-01-02 00:00:00"),
            (datetime.date(1970, 1, 2), "DATE", "1970-01-02"),
            (datetime.date(1000, 1, 2), "DATE", "1000-01-02"),
            (datetime.datetime(1970, 1, 1, 1, 0), None, "1970-01-01 01:00:00"),
            (datetime.datetime(1000, 1, 1, 1, 0), None, "1000-01-01 01:00:00"),
            (datetime.timedelta(), None, "00:00:00"),
            (datetime.timedelta(hours=23, minutes=59, seconds=59), None, "23:59:59"),
            (decimal.Decimal(5), None, 5),
            (b"bytes", "BYTES", "Ynl0ZXM="),
            (b"\x00\x01", "INTEGER", 1),
            (None, "BYTES", None),
        ],
    )
    def test_convert_type(self, value, schema_type, expected):
        op = MySQLToGCSOperator(
            task_id=TASK_ID, mysql_conn_id=MYSQL_CONN_ID, sql=SQL, bucket=BUCKET, filename=JSON_FILENAME
        )
        assert op.convert_type(value, schema_type) == expected

    @mock.patch("airflow.providers.google.cloud.transfers.mysql_to_gcs.MySqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.sql_to_gcs.GCSHook")
    def test_exec_success_json(self, gcs_hook_mock_class, mysql_hook_mock_class):
        """Test successful run of execute function for JSON"""
        op = MySQLToGCSOperator(
            task_id=TASK_ID, mysql_conn_id=MYSQL_CONN_ID, sql=SQL, bucket=BUCKET, filename=JSON_FILENAME
        )

        mysql_hook_mock = mysql_hook_mock_class.return_value
        mysql_hook_mock.get_conn().cursor().__iter__.return_value = iter(ROWS)
        mysql_hook_mock.get_conn().cursor().description = CURSOR_DESCRIPTION

        gcs_hook_mock = gcs_hook_mock_class.return_value

        def _assert_upload(bucket, obj, tmp_filename, mime_type=None, gzip=False, metadata=None):
            assert bucket == BUCKET
            assert JSON_FILENAME.format(0) == obj
            assert mime_type == "application/json"
            assert not gzip
            with open(tmp_filename, "rb") as file:
                assert b"".join(NDJSON_LINES) == file.read()

        gcs_hook_mock.upload.side_effect = _assert_upload

        op.execute(None)

        mysql_hook_mock_class.assert_called_once_with(mysql_conn_id=MYSQL_CONN_ID)
        mysql_hook_mock.get_conn().cursor().execute.assert_called_once_with(SQL)

    @mock.patch("airflow.providers.google.cloud.transfers.mysql_to_gcs.MySqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.sql_to_gcs.GCSHook")
    def test_exec_success_csv(self, gcs_hook_mock_class, mysql_hook_mock_class):
        """Test successful run of execute function for CSV"""
        op = MySQLToGCSOperator(
            task_id=TASK_ID,
            mysql_conn_id=MYSQL_CONN_ID,
            sql=SQL,
            export_format="CSV",
            bucket=BUCKET,
            filename=CSV_FILENAME,
        )

        mysql_hook_mock = mysql_hook_mock_class.return_value
        mysql_hook_mock.get_conn().cursor().__iter__.return_value = iter(ROWS)
        mysql_hook_mock.get_conn().cursor().description = CURSOR_DESCRIPTION

        gcs_hook_mock = gcs_hook_mock_class.return_value

        def _assert_upload(bucket, obj, tmp_filename, mime_type=None, gzip=False, metadata=None):
            assert bucket == BUCKET
            assert CSV_FILENAME.format(0) == obj
            assert mime_type == "text/csv"
            assert not gzip
            with open(tmp_filename, "rb") as file:
                assert b"".join(CSV_LINES) == file.read()

        gcs_hook_mock.upload.side_effect = _assert_upload

        op.execute(None)

        mysql_hook_mock_class.assert_called_once_with(mysql_conn_id=MYSQL_CONN_ID)
        mysql_hook_mock.get_conn().cursor().execute.assert_called_once_with(SQL)

    @mock.patch("airflow.providers.google.cloud.transfers.mysql_to_gcs.MySqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.sql_to_gcs.GCSHook")
    def test_exec_success_csv_ensure_utc(self, gcs_hook_mock_class, mysql_hook_mock_class):
        """Test successful run of execute function for CSV"""
        op = MySQLToGCSOperator(
            task_id=TASK_ID,
            mysql_conn_id=MYSQL_CONN_ID,
            sql=SQL,
            export_format="CSV",
            bucket=BUCKET,
            filename=CSV_FILENAME,
            ensure_utc=True,
        )

        mysql_hook_mock = mysql_hook_mock_class.return_value
        mysql_hook_mock.get_conn().cursor().__iter__.return_value = iter(ROWS)
        mysql_hook_mock.get_conn().cursor().description = CURSOR_DESCRIPTION

        gcs_hook_mock = gcs_hook_mock_class.return_value

        def _assert_upload(bucket, obj, tmp_filename, mime_type=None, gzip=False, metadata=None):
            assert bucket == BUCKET
            assert CSV_FILENAME.format(0) == obj
            assert mime_type == "text/csv"
            assert not gzip
            with open(tmp_filename, "rb") as file:
                assert b"".join(CSV_LINES) == file.read()

        gcs_hook_mock.upload.side_effect = _assert_upload

        op.execute(None)

        mysql_hook_mock_class.assert_called_once_with(mysql_conn_id=MYSQL_CONN_ID)
        mysql_hook_mock.get_conn().cursor().execute.assert_has_calls([mock.call(TZ_QUERY), mock.call(SQL)])

    @mock.patch("airflow.providers.google.cloud.transfers.mysql_to_gcs.MySqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.sql_to_gcs.GCSHook")
    def test_exec_success_csv_with_delimiter(self, gcs_hook_mock_class, mysql_hook_mock_class):
        """Test successful run of execute function for CSV with a field delimiter"""
        op = MySQLToGCSOperator(
            task_id=TASK_ID,
            mysql_conn_id=MYSQL_CONN_ID,
            sql=SQL,
            export_format="csv",
            field_delimiter="|",
            bucket=BUCKET,
            filename=CSV_FILENAME,
        )

        mysql_hook_mock = mysql_hook_mock_class.return_value
        mysql_hook_mock.get_conn().cursor().__iter__.return_value = iter(ROWS)
        mysql_hook_mock.get_conn().cursor().description = CURSOR_DESCRIPTION

        gcs_hook_mock = gcs_hook_mock_class.return_value

        def _assert_upload(bucket, obj, tmp_filename, mime_type=None, gzip=False, metadata=None):
            assert bucket == BUCKET
            assert CSV_FILENAME.format(0) == obj
            assert mime_type == "text/csv"
            assert not gzip
            with open(tmp_filename, "rb") as file:
                assert b"".join(CSV_LINES_PIPE_DELIMITED) == file.read()

        gcs_hook_mock.upload.side_effect = _assert_upload

        op.execute(None)

        mysql_hook_mock_class.assert_called_once_with(mysql_conn_id=MYSQL_CONN_ID)
        mysql_hook_mock.get_conn().cursor().execute.assert_called_once_with(SQL)

    @mock.patch("airflow.providers.google.cloud.transfers.mysql_to_gcs.MySqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.sql_to_gcs.GCSHook")
    def test_file_splitting(self, gcs_hook_mock_class, mysql_hook_mock_class):
        """Test that ndjson is split by approx_max_file_size_bytes param."""
        mysql_hook_mock = mysql_hook_mock_class.return_value
        mysql_hook_mock.get_conn().cursor().__iter__.return_value = iter(ROWS)
        mysql_hook_mock.get_conn().cursor().description = CURSOR_DESCRIPTION

        gcs_hook_mock = gcs_hook_mock_class.return_value
        expected_upload = {
            JSON_FILENAME.format(0): b"".join(NDJSON_LINES[:2]),
            JSON_FILENAME.format(1): NDJSON_LINES[2],
        }

        def _assert_upload(bucket, obj, tmp_filename, mime_type=None, gzip=False, metadata=None):
            assert bucket == BUCKET
            assert mime_type == "application/json"
            assert not gzip
            with open(tmp_filename, "rb") as file:
                assert expected_upload[obj] == file.read()

        gcs_hook_mock.upload.side_effect = _assert_upload

        op = MySQLToGCSOperator(
            task_id=TASK_ID,
            sql=SQL,
            bucket=BUCKET,
            filename=JSON_FILENAME,
            approx_max_file_size_bytes=len(expected_upload[JSON_FILENAME.format(0)]),
        )
        op.execute(None)

    @mock.patch("airflow.providers.google.cloud.transfers.mysql_to_gcs.MySqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.sql_to_gcs.GCSHook")
    def test_schema_file(self, gcs_hook_mock_class, mysql_hook_mock_class):
        """Test writing schema files."""
        mysql_hook_mock = mysql_hook_mock_class.return_value
        mysql_hook_mock.get_conn().cursor().__iter__.return_value = iter(ROWS)
        mysql_hook_mock.get_conn().cursor().description = CURSOR_DESCRIPTION

        gcs_hook_mock = gcs_hook_mock_class.return_value

        def _assert_upload(bucket, obj, tmp_filename, mime_type, gzip, metadata=None):
            if obj == SCHEMA_FILENAME:
                assert not gzip
                with open(tmp_filename, "rb") as file:
                    assert b"".join(SCHEMA_JSON) == file.read()

        gcs_hook_mock.upload.side_effect = _assert_upload

        op = MySQLToGCSOperator(
            task_id=TASK_ID, sql=SQL, bucket=BUCKET, filename=JSON_FILENAME, schema_filename=SCHEMA_FILENAME
        )
        op.execute(None)

        # once for the file and once for the schema
        assert gcs_hook_mock.upload.call_count == 2

    @mock.patch("airflow.providers.google.cloud.transfers.mysql_to_gcs.MySqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.sql_to_gcs.GCSHook")
    def test_schema_file_with_custom_schema(self, gcs_hook_mock_class, mysql_hook_mock_class):
        """Test writing schema files with customized schema"""
        mysql_hook_mock = mysql_hook_mock_class.return_value
        mysql_hook_mock.get_conn().cursor().__iter__.return_value = iter(ROWS)
        mysql_hook_mock.get_conn().cursor().description = CURSOR_DESCRIPTION

        gcs_hook_mock = gcs_hook_mock_class.return_value

        def _assert_upload(bucket, obj, tmp_filename, mime_type, gzip, metadata=None):
            if obj == SCHEMA_FILENAME:
                assert not gzip
                with open(tmp_filename, "rb") as file:
                    assert b"".join(CUSTOM_SCHEMA_JSON) == file.read()

        gcs_hook_mock.upload.side_effect = _assert_upload

        op = MySQLToGCSOperator(
            task_id=TASK_ID,
            sql=SQL,
            bucket=BUCKET,
            filename=JSON_FILENAME,
            schema_filename=SCHEMA_FILENAME,
            schema=SCHEMA,
        )
        op.execute(None)

        # once for the file and once for the schema
        assert gcs_hook_mock.upload.call_count == 2

    @mock.patch("airflow.providers.google.cloud.transfers.mysql_to_gcs.MySqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.sql_to_gcs.GCSHook")
    def test_query_with_error(self, mock_gcs_hook, mock_mysql_hook):
        mock_mysql_hook.return_value.get_conn.return_value.cursor.return_value.execute.side_effect = (
            ProgrammingError
        )
        op = MySQLToGCSOperator(
            task_id=TASK_ID, sql=SQL, bucket=BUCKET, filename=JSON_FILENAME, schema_filename=SCHEMA_FILENAME
        )
        with pytest.raises(ProgrammingError):
            op.query()

    @mock.patch("airflow.providers.google.cloud.transfers.mysql_to_gcs.MySqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.sql_to_gcs.GCSHook")
    def test_execute_with_query_error(self, mock_gcs_hook, mock_mysql_hook):
        mock_mysql_hook.return_value.get_conn.return_value.cursor.return_value.execute.side_effect = (
            ProgrammingError
        )
        op = MySQLToGCSOperator(
            task_id=TASK_ID, sql=SQL, bucket=BUCKET, filename=JSON_FILENAME, schema_filename=SCHEMA_FILENAME
        )
        with pytest.raises(ProgrammingError):
            op.execute(None)

    @pytest.mark.parametrize(
        "connection_port, default_port, expected_port",
        [(None, 4321, 4321), (1234, None, 1234), (1234, 4321, 1234)],
    )
    def test_execute_openlineage_events(self, connection_port, default_port, expected_port):
        class DBApiHookForTests(DbApiHook):
            conn_name_attr = "sql_default"
            get_conn = mock.MagicMock(name="conn")
            get_connection = mock.MagicMock()

            def get_openlineage_database_info(self, connection):
                from airflow.providers.openlineage.sqlparser import DatabaseInfo

                return DatabaseInfo(
                    scheme="sqlscheme",
                    authority=DbApiHook.get_openlineage_authority_part(connection, default_port=default_port),
                )

        dbapi_hook = DBApiHookForTests()

        class MySQLToGCSOperatorForTest(MySQLToGCSOperator):
            @property
            def db_hook(self):
                return dbapi_hook

        sql = """SELECT a,b,c from my_db.my_table"""
        op = MySQLToGCSOperatorForTest(task_id=TASK_ID, sql=sql, bucket="bucket", filename="dir/file{}.csv")
        DB_SCHEMA_NAME = "PUBLIC"
        rows = [
            (DB_SCHEMA_NAME, "popular_orders_day_of_week", "order_day_of_week", 1, "varchar"),
            (DB_SCHEMA_NAME, "popular_orders_day_of_week", "order_placed_on", 2, "timestamp"),
            (DB_SCHEMA_NAME, "popular_orders_day_of_week", "orders_placed", 3, "int4"),
        ]
        dbapi_hook.get_connection.return_value = Connection(
            conn_id="sql_default", conn_type="mysql", host="host", port=connection_port
        )
        dbapi_hook.get_conn.return_value.cursor.return_value.fetchall.side_effect = [rows, []]

        lineage = op.get_openlineage_facets_on_start()
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].namespace == f"sqlscheme://host:{expected_port}"
        assert lineage.inputs[0].name == "PUBLIC.popular_orders_day_of_week"
        assert len(lineage.inputs[0].facets) == 1
        assert lineage.inputs[0].facets["schema"].fields == [
            SchemaDatasetFacetFields(name="order_day_of_week", type="varchar"),
            SchemaDatasetFacetFields(name="order_placed_on", type="timestamp"),
            SchemaDatasetFacetFields(name="orders_placed", type="int4"),
        ]
        assert lineage.outputs == [
            OutputDataset(
                namespace="gs://bucket",
                name="dir",
            )
        ]

        assert len(lineage.job_facets) == 1
        assert lineage.job_facets["sql"].query == sql
        assert lineage.run_facets == {}
